from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DfGeolocationConfig(AppConfig):
    name = "df_geolocation"
    verbose_name = _("Django DF GEO Location")



from django.db import models
from django.utils.translation import gettext_lazy as _
from django_countries.fields import CountryField

from .querysets import PositionQuerySet


class AddressMixin(models.Model):
    formatted = models.CharField(max_length=256, null=True, blank=True)
    street = models.CharField(
        _("Street Address"), max_length=255, blank=True, null=True
    )
    city = models.CharField(_("City"), max_length=100, blank=True, null=True)
    postcode = models.CharField(_("Postcode"), max_length=8, null=True, blank=True)
    country = CountryField(_("Country"), default="NL")

    def __str__(self) -> str:
        return f"{self.street}, {self.postcode} {self.city}, {self.country}"

    class Meta:
        abstract = True


class PositionMixin(models.Model):
    lat = models.FloatField(_("Latitude"))
    lon = models.FloatField(_("Longitude"))

    objects = PositionQuerySet.as_manager()

    class Meta:
        abstract = True


class Address(AddressMixin, PositionMixin):
    class Meta:
        abstract = True



from typing import Any

from django.db import models

from .functions import EarthDistance, LLToEarth


class PositionQuerySet(models.QuerySet):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.lat_field = getattr(self.model, "position_field_name", "") + "lat"
        self.lon_field = getattr(self.model, "position_field_name", "") + "lon"

    # This was an old implementation that did not seem to work
    # def in_range(self, lat, lon, meters):
    #     return self.annotate(
    #         earthbox=EarthBox(LLToEarth(lat, lon), meters),
    #     ).filter(earthbox__in_georange=LLToEarth(self.lat_field, self.lon_field))

    def in_range(
        self, lat: float, lon: float, meters: float = 0.0, miles: float = 0.0
    ) -> models.QuerySet:
        meters = meters or (miles * 1609.34)
        return self.annotate_range(lat, lon).filter(range__lte=meters)

    def annotate_range(self, lat: float, lon: float) -> models.QuerySet:
        return self.annotate(
            range=EarthDistance(
                LLToEarth(self.lat_field, self.lon_field), LLToEarth(lat, lon)
            )
        )



from django_countries.serializer_fields import CountryField
from hashid_field.rest import HashidSerializerCharField
from rest_framework import serializers


class AddressSerializer(serializers.ModelSerializer):
    id = HashidSerializerCharField(read_only=True)
    country = CountryField(required=False)

    class Meta:
        read_only_fields = ("id",)
        fields = read_only_fields + (
            "formatted",
            "street",
            "city",
            "postcode",
            "country",
            "notes",
            "is_current",
            "lat",
            "lon",
        )


class PositionSerializer(serializers.Serializer):
    lat = serializers.FloatField()
    lon = serializers.FloatField()



from typing import Any

from django.db.models import Field, Func, Lookup, fields


class LLToEarth(Func):
    function = "ll_to_earth"
    arg_joiner = ", "
    arity = 2  # The number of arguments the function accepts.

    def __init__(
        self, *expressions: Any, output_field: Any = None, **extra: Any
    ) -> None:
        if output_field is None:
            output_field = fields.Field()
        super().__init__(*expressions, output_field=output_field, **extra)


class EarthBox(LLToEarth):
    function = "earth_box"
    arg_joiner = ", "


class EarthDistance(LLToEarth):
    function = "earth_distance"
    arg_joiner = ", "


@Field.register_lookup
class Near(Lookup):
    lookup_name = "in_georange"
    operator = "@>"

    def as_sql(self, compiler: Any, connection: Any) -> tuple[str, Any]:
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s @> %s" % (lhs, rhs), params
